local tprequests = {}
local tpcooldown = {}
local tpstate = {}

local paceframe = {}
local pacetime = {}

local cv_tpcooldowntime = CV_RegisterVar({
	name = "tpcooldowntime",
	defaultvalue = "30",
	flags = CV_NETVAR,
	PossibleValue = CV_Natural,
})

local cv_tptimeout = CV_RegisterVar({
	name = "tptimeout",
	defaultvalue = "18",
	flags = CV_NETVAR,
	PossibleValue = CV_Natural,
})

local cv_visor = CV_RegisterVar({
	name = "visor",
	defaultvalue = "Off",
	flags = 0,
	PossibleValue = CV_OnOff,
})

local WALKSPEED = 9 << FRACBITS

if not tofixed then
	-- 2.2.13 compatibility
	rawset(_G, "tofixed", function(n)
		local delim = n:find('.', 0, true)
		if delim then
			local exp = tonumber(n:sub(1, delim-1))
			local frac = tonumber(n:sub(delim+1))
			if exp == nil or frac == nil then
				return nil
			end
			return (exp << FRACBITS) | ((frac << FRACBITS) / (10 ^ (#n-delim)))
		end
		local val = tonumber(n)
		if val == nil then
			return nil
		end
		return val << FRACBITS
	end)
end

COM_AddCommand("pace", function (player, speed)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to change movement pace!")
		return
	end
	if speed == nil then
		if player.normalspeed == skins[player.mo.skin].normalspeed then
			CONS_Printf(player, "You are currently moving at default speed (" .. (player.normalspeed >> FRACBITS) .. " FU/tic)")
		elseif player.normalspeed == skins[player.mo.skin].runspeed - FRACUNIT then
			CONS_Printf(player, "You are currently moving at jogging speed (" .. (player.normalspeed >> FRACBITS) .. " FU/tic)")
		elseif player.normalspeed == WALKSPEED then
			CONS_Printf(player, "You are currently moving at walking speed (10 FU/tic)")
		else
			CONS_Printf(player, "You are currently moving at " .. (player.normalspeed >> FRACBITS) .. " FU/tic")
		end
		return
	end
	if speed:lower() == "walk" then
		player.normalspeed = WALKSPEED
		return
	elseif speed:lower() == "jog" then
		-- subtract one to make sure we don't go into running animation
		player.normalspeed = skins[player.mo.skin].runspeed - FRACUNIT
		return
	elseif speed:lower() == "default" or speed:lower() == "normal" then
		player.normalspeed = skins[player.mo.skin].normalspeed
		return
	elseif speed:lower() == "super" then
		player.normalspeed = skins[player.mo.skin].normalspeed * (player.powers[pw_super] > 0 and 1 or 2)
		return
	end
	local nspeed = tonumber(speed)
	if nspeed ~= nil and nspeed > 0 and nspeed <= 200 then
		player.normalspeed = nspeed << FRACBITS
	else
		CONS_Printf(player, speed .. ": not a valid value")
	end
end)

COM_AddCommand("jumpheight", function (player, speed)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to change jump height!")
		return
	end
	if speed == nil then
		CONS_Printf(player, "Your current jump height is " .. (player.jumpfactor >> FRACBITS) .. "." .. (((player.jumpfactor * 10) >> FRACBITS) % 10))
		return
	end
	if speed:lower() == "default" or speed:lower() == "normal" then
		player.jumpfactor = skins[player.mo.skin].jumpfactor
		return
	end
	local nspeed = tofixed(speed)
	if nspeed ~= nil and nspeed > 0 and nspeed <= 5 * FRACUNIT then
		player.jumpfactor = nspeed
	else
		CONS_Printf(player, speed .. ": not a valid value")
	end
end)

COM_AddCommand("knockback", function (player, speed)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to be knocked back!")
		return
	end
	P_DoPlayerPain(player)
	S_StartSound(player.mo, sfx_shldls)
end)

local fadevalue = 0
local targettime = nil

addHook("MapLoad", function (mapnumber)
	tprequests = {}
	tpcooldown = {}
	targettime = nil
	tpstate = {}
end)

addHook("NetVars", function (network)
	tprequests = network(tprequests)
	tpcooldown = network(tpcooldown)
	targettime = network(targettime)
end)

local function DoTeleport(player, target)
	S_StartSound(player.mo, sfx_s3k75)
	player.mo.target = target.mo
	tpstate[#player] = 1
end

local function FindPlayerByNameOrNumber(name)
	local playernum = tonumber(name)
	if playernum ~= nil and playernum >= 0 and playernum <= 32 then
		local player = players[playernum]
		if not (player and player.valid) then
			return nil
		end
		return  player
	end
	for player in players.iterate do
		if player and player.valid and player.name == name then
			return player
		end
	end
	return nil
end

COM_AddCommand("tpto", function (player, target, force)
	if not player.mo then
		CONS_Printf(player, "You must be present in a map to teleport!")
		return
	end
	if tpcooldown[#player] and leveltime < tpcooldown[#player] then
		CONS_Printf(player, "You must wait " .. ((tpcooldown[#player] - leveltime) / TICRATE) .. " seconds before you can request a teleport again!")
		return
	end
	if target == nil then
		CONS_Printf(player, "tpto <player>: teleports to a player")
		return
	end
	local t = FindPlayerByNameOrNumber(target)
	if not t then
		CONS_Printf(player, "Player " .. target .. " is unknown or invalid")
		return
	end
	target = t
	local playernum = #target

	if not (target and target.valid) then
		CONS_Printf(player, "Player " .. playernum .. " is unknown or invalid")
		return
	end
	if not target.mo then
		CONS_Printf(player, "Player " .. playernum .. " is not present on the map")
		return
	end
	if playernum == #player then
		CONS_Printf(player, "You can't teleport to yourself!")
		return
	end
	if force == "-f" or force == "-force" then
		if not IsPlayerAdmin(player) then
			CONS_Printf(player, "Only the server or remote admin can force teleport")
			return
		end
		CONS_Printf(player, "Teleporting to " .. target.name)
		DoTeleport(player, target)
		return
	end
	if tprequests[playernum] then
		CONS_Printf(player, "Someone has already requested a teleport to this player, try again later")
		return
	end
	tprequests[playernum] = { #player, leveltime + cv_tptimeout.value * TICRATE }
	CONS_Printf(player, "Requested teleport to " .. target.name)
	CONS_Printf(target, player.name .. " is requesting to teleport to you")
	S_StartSound(nil, sfx_strpst, target)
end)

COM_AddCommand("tpyes", function (player)
	if not tprequests[#player] then
		CONS_Printf(player, "Nobody has requested a teleport to you")
		return
	end
	if not player.mo then
		CONS_Printf(player, "You cannot accept a teleport when you don't exist in the map!")
		return
	end

	local source = players[tprequests[#player][1]]
	if not (source and source.valid and source.mo) then
		CONS_Printf(player, "Player requesting a teleport no longer exists")
		tprequests[#player] = nil
		return
	end
	CONS_Printf(player, "Accepted teleport from " .. source.name)
	CONS_Printf(source, player.name .. " has accepted your teleport request")
	tprequests[#player] = nil
	DoTeleport(source, player)
end)

COM_AddCommand("tpno", function (player)
	if not tprequests[#player] then
		CONS_Printf(player, "Nobody has requested a teleport to you")
		return
	end

	local source = players[tprequests[#player][1]]
	if not (source and source.valid and source.mo) then
		CONS_Printf(player, "Player requesting a teleport no longer exists")
		tprequests[#player] = nil
		return
	end
	CONS_Printf(player, "Rejected teleport from " .. source.name)
	CONS_Printf(source, player.name .. " has rejected your teleport request")
	tpcooldown[#source] = leveltime + cv_tpcooldowntime.value * TICRATE
	tprequests[#player] = nil
end)

local KEY_ENTER = 13
local KEY_ESCAPE = 27

local KEY_Y = 121
local KEY_N = 110

local chat_on = false

addHook("KeyDown", function (keyevent)
	if chat_on then
		if keyevent.num == KEY_ENTER or keyevent.num == KEY_ESCAPE then
			chat_on = false
		end
		return
	end
	if not (consoleplayer and consoleplayer.valid) then
		return
	end
	if tprequests[#consoleplayer] then
		if keyevent.num == KEY_Y then
			COM_BufAddText(nil, "tpyes")
			return true
		elseif keyevent.num == KEY_N then
			COM_BufAddText(nil, "tpno")
			return true
		end
	end
	if keyevent.num == input.gameControlToKeyNum(GC_TALKKEY) or keyevent.num == input.gameControlToKeyNum(GC_TEAMKEY)
		chat_on = true
	end
end)

addHook("PlayerQuit", function (player)
	-- clear all tp requests so we don't have any lingering teleports
	tprequests[#player] = nil
	for i = 1,#players do
		if tprequests[i] ~= nil and tprequests[i][1] == #player then
			tprequests[i] = nil
		end
	end
end)

addHook("ThinkFrame", function ()
	if consoleplayer and tprequests[#consoleplayer] ~= nil and tprequests[#consoleplayer][2] == leveltime then
		COM_BufAddText(players[#consoleplayer], "tpno")
	end
	if fadevalue >= 10 then
		if not consoleplayer.flashcount then
			-- keep flashing the palette to ensure it doesn't go away
			P_FlashPal(consoleplayer, 6, 1)
		end
	end
end)

hud.add(function (v, stplyr, cam)
	if fadevalue ~= 0 and fadevalue ~= 20 then
		v.fadeScreen(6, min(fadevalue, 20-fadevalue))
	end
	if cv_visor.value then
		if fadevalue == 0 then
			S_StartSound(nil, sfx_s1c9)
		end
		if fadevalue < 20 then
			fadevalue = fadevalue + 1
		end
	else
		if fadevalue == 20 then
			S_StartSound(nil, sfx_s1c9)
		end
		if fadevalue > 0 then
			fadevalue = fadevalue - 1
		end
	end
	if targettime and targettime >= leveltime then
		local remain = targettime - leveltime
		local map = V_YELLOWMAP
		if remain <= 5 * TICRATE then
			if remain % TICRATE >= 30 then
				if remain % TICRATE == TICRATE - 1 then
					S_StartSound(nil, sfx_ding)
				end
				map = V_REDMAP
			elseif remain % TICRATE >= 25 then
				map = V_BROWNMAP
			elseif remain % TICRATE >= 20 then
				map = V_ORANGEMAP
			elseif remain % TICRATE >= 15 then
				map = V_PERIDOTMAP
			end
		end
		v.drawString(160, 140, (remain + TICRATE) / TICRATE, map|V_SNAPTOBOTTOM, "center")
	end

	if not tprequests[#consoleplayer] then
		return
	end

	local target = players[tprequests[#consoleplayer][1]]
	if not (target and target.valid) then
		return
	end

	local f = V_YELLOWMAP | V_SNAPTOTOP
	v.drawString(160, 60, target.name .. " want to teleport to you!", f, "small-center")
	v.drawString(160, 65, "Do you wish to accept? (Y/N)", f, "small-center")
end)

COM_AddCommand("shield", function (player, shield)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to use shields!")
		return
	end
	if shield == nil then
		CONS_Printf(player, "Set or remove your shield")
		CONS_Printf(player, "Available shields: none, pity, whirlwind, armageddon, pink, elemental, attract, flameaura, bubblewrap, thundercoin, force")
		return
	end
	shield = shield:lower()
	local sound
	if shield == "none" then
		player.powers[pw_shield] = 0
		P_RemoveShield(player)
		S_StartSound(player.mo, sfx_shldls)
		return
	elseif shield == "pity" then
		player.powers[pw_shield] = SH_PITY
		sound = sfx_shield
	elseif shield == "whirlwind" then
		player.powers[pw_shield] = SH_WHIRLWIND
		sound = sfx_wirlsg
	elseif shield == "armageddon" then
		player.powers[pw_shield] = SH_ARMAGEDDON
		sound = sfx_armasg
	elseif shield == "pink" then
		player.powers[pw_shield] = SH_PINK
		sound = sfx_shield
	elseif shield == "elemental" then
		player.powers[pw_shield] = SH_ELEMENTAL
		sound = sfx_elemsg
	elseif shield == "attract" then
		player.powers[pw_shield] = SH_ATTRACT
		sound = sfx_attrsg
	elseif shield == "flameaura" then
		player.powers[pw_shield] = SH_FLAMEAURA
		sound = sfx_s3k3e
	elseif shield == "bubblewrap" then
		player.powers[pw_shield] = SH_BUBBLEWRAP
		sound = sfx_s3k3f
	elseif shield == "thundercoin" then
		player.powers[pw_shield] = SH_THUNDERCOIN
		sound = sfx_s3k41
	elseif shield == "force" then
		player.powers[pw_shield] = SH_FORCE|1
		sound = sfx_forcsg
	else
		CONS_Printf(player, "Unknown shield " .. shield)
		return
	end
	P_SpawnShieldOrb(player)
	S_StartSound(player.mo, sound)
end)

COM_AddCommand("phase", function (player)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to phase!")
		return
	end
	if player.pflags & PF_NOCLIP then
		-- check if it's safe to turn off phasing here
		local subsector = R_PointInSubsectorOrNil(player.mo.x, player.mo.y)
		if R_PointInSubsectorOrNil(player.mo.x, player.mo.y) == nil then
			CONS_Printf(player, "Can't unphase here: you are out of bounds!")
			return
		end
		if player.mo.ceilingz - player.mo.floorz < player.mo.height then
			CONS_Printf(player, "Can't unphase here: not enough headroom to fit!")
			return
		end
	end
	player.pflags = player.pflags ^^ (PF_NOCLIP|PF_GODMODE)
	player.powers[pw_invulnerability] = 0xffff * (player.pflags & PF_NOCLIP)
	player.mo.ht_trans = 0
	if player.pflags & PF_NOCLIP then
		player.mo.ht_trans = 30
	end
	S_StartSound(player.mo, sfx_ngskid)
end)

local MINSCALE = FRACUNIT / 10
local MAXSCALE = FRACUNIT * 20

COM_AddCommand("scale", function (player, scale)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to scale!")
		return
	end
	if scale == nil then
		CONS_Printf(player, "scale <number>: set your character's size")
		return
	end
	local n = tofixed(scale)
	if n == nil then
		CONS_Printf(player, scale .. ": not a valid value")
		return
	end
	if n < MINSCALE then
		CONS_Printf(player, "Can't scale to " .. scale .. ": you cannot become smaller than 0.1!")
		return
	end
	if n > MAXSCALE then
		CONS_Printf(player, "Can't scale to " .. scale .. ": you cannot become larger than 20!")
		return
	end
	-- make sure the player fits the room
	if not (player.pflags & PF_NOCLIP) and player.mo.ceilingz - player.mo.floorz < FixedMul(n, FixedDiv(player.mo.height, player.mo.scale)) then
		CONS_Printf(player, "Can't scale to " .. scale .. ": not enough headroom to fit!")
		return
	end
	player.mo.destscale = n
	CONS_Printf(player, "Scaling to " .. scale)
end)

COM_AddCommand("cloak", function (player, trans)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to change your transparency!")
		return
	end
	if player.pflags & PF_NOCLIP then
		CONS_Printf(player, "You cannot cloak while phased.")
		return
	end
	if trans == nil then
		CONS_Printf(player, "cloak <0-10>: set your character's transparency")
		return
	end
	local n = tonumber(trans)
	if n == nil then
		CONS_Printf(player, trans .. ": not a valid value")
		return
	end
	if n < 0 or n > 10 then
		CONS_Printf(player, "Transparency level must be between 0 and 10!")
		return
	end
	player.mo.ht_trans = n * 6
	S_StartSound(player.mo, sfx_ngskid)
end)

COM_AddCommand("super", function (player)
	if not player.mo then
		CONS_Printf(player, "You need to be present in the game to go super!")
		return
	end
	if not (skins[player.skin].flags & SF_SUPER)
		CONS_Printf(player, skins[player.skin].realname .. " cannot go super.")
		return
	end
	if player.powers[pw_super] then
		player.rings = 0
		return
	end
	player.rings = 9999
	P_DoSuperTransformation(player)
end)

COM_AddCommand("countdown", function (player, count)
	if targettime ~= nil and targettime >= leveltime then
		CONS_Printf(player, "A countdown is already ongoing, cancel it with stopcountdown first.")
		return
	end
	if count == nil then
		CONS_Printf(player, "You didn't specify a time!")
		return
	end
	local n = tonumber(count)
	if n == nil then
		CONS_Printf(player, count .. " is not a valid value")
		return
	end
	targettime = leveltime + n * TICRATE - 1
end, COM_ADMIN)

COM_AddCommand("stopcountdown", function (player)
	if targettime ~= nil and targettime >= leveltime then
		targettime = nil
	else
		CONS_Printf(player, "There is no countdown active at the moment.")
		return
	end
end, COM_ADMIN)

local flipping = {}

COM_AddCommand("flip", function (player)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to flip!")
		return
	end
	player.mo.flags2 = player.mo.flags2 ^^ MF2_OBJECTFLIP
	player.mo.spriteroll = ANGLE_180
	if player.pflags & PF_FLIPCAM then
		player.viewrollangle = ANGLE_180
	end
	S_StartSound(player.mo, sfx_GDFLIP)
	flipping[#player] = true
end)

COM_AddCommand("thrust", function (player, forward, up)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to thrust!")
		return
	end

	if forward == nil then
		CONS_Printf(player, "thrust <forward> [up]: thrusts you forward and optionally up")
		return
	end
	local f = tonumber(forward)
	if f == nil or f < 0 or f > 100 then
		CONS_Printf(player, forward .. ": not a valid value")
		return
	end
	P_Thrust(player.mo, player.mo.angle, f << FRACBITS)
	S_StartSound(player.mo, sfx_GDTRST)
	if up ~= nil then
		local u = tonumber(up)
		if u == nil or u < -100 or u > 100 then
			CONS_Printf(player, up .. ": not a valid value")
			return
		end
		if player.mo.flags2 & MF2_OBJECTFLIP then
			u = -u
		end
		player.mo.momz = player.mo.momz + (u << FRACBITS)
	end
	local mo = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SCALEGHOST)
	if trans then
		mo.state = S_SCALEGHOST1 + trans - 1
	end
	mo.destscale = mo.scale
	mo.angle = player.drawangle
	mo.pitch = player.mo.pitch
	mo.roll = player.mo.roll
	mo.spriteroll = player.mo.spriteroll
	mo.cusval = player.mo.sprite2
	mo.cvmem = (player.mo.frame & FF_FRAMEMASK)
	mo.skin = player.mo.skin
	mo.color = SKINCOLOR_SUPERPURPLE1
	mo.colorized = true
end)

local MAXWEIGHT = 50 * FRACUNIT

COM_AddCommand("weight", function (player, value)
	if not (player.mo and player.mo.valid) then
		CONS_Printf(player, "You need to be present in the game to set your weight!")
		return
	end
	if value == nil then
		CONS_Printf(player, "weight <number>: set your own gravity")
		if player.mo.ht_weight == nil then
			CONS_Printf(player, "Your current weight is 1")
		else
			CONS_Printf(player, "Your current weight is " .. player.mo.ht_weight)
		end
		return
	end
	local n = tofixed(value)
	if n == nil or n < 0 or n > MAXWEIGHT then
		CONS_Printf(player, value .. ": not a valid value")
		return
	end
	if n == 1 then
		player.mo.ht_weight = nil
		player.mo.flags = player.mo.flags & MF_NOGRAVITY
		return
	end
	player.mo.ht_weight = n
	player.mo.flags = player.mo.flags | MF_NOGRAVITY
end)

-- visual only, don't bother syncing
rawset(_G, "ht_playertrans", { })
local MAXPLAYERTRANS = 35
local playerscaling = { }

for i = 0,#players do
	ht_playertrans[i] = 0
	playerscaling[i] = false
end

addHook("PlayerSpawn", function (player)
	ht_playertrans[#player] = 0
	playerscaling[#player] = false
	emeralds = 0x7f
	tpstate[#player] = nil
	if player.mo then
		player.normalspeed = skins[player.mo.skin].normalspeed
		player.jumpfactor = skins[player.mo.skin].jumpfactor
	else
		player.normalspeed = skins[player.skin].normalspeed
		player.jumpfactor = skins[player.skin].jumpfactor
	end
end)

addHook("PlayerThink", function (player)
	if not player.mo then return end
	local mobj = player.mo
	if mobj.ht_weight ~= nil and not P_IsObjectOnGround(mobj) then
		local g = P_GetMobjGravity(mobj)
		if mobj.momz == 0 then
			-- compensate for a special case in the engine
			g = g << 1
		end
		mobj.momz = mobj.momz + FixedMul(g, mobj.ht_weight)
	end
end, MT_PLAYER)

local function HandlePlayerOverlay(player, field, frame, time, skincolor, condition)
	local mo = player.mo[field]
	if mo and mo.valid then
		if frame[#player] == nil then
			frame[#player] = 0
		else
			frame[#player] = frame[#player] + 1
		end
		if frame[#player] < 3*4 then
			mo.color = skincolor + frame[#player] / 3
		elseif frame[#player] > 3*7 then
			mo.color = skincolor
			frame[#player] = 0
		else
			mo.color = skincolor + 4 - (frame[#player] - 3*4) / 3
		end
		P_SetMobjStateNF(mo, player.mo.state)
		mo.destscale = mo.scale
		mo.angle = player.drawangle
		mo.pitch = player.mo.pitch
		mo.roll = player.mo.roll
		mo.spriteroll = player.mo.spriteroll
		mo.sprite = player.mo.sprite
		mo.sprite2 = player.mo.sprite2
		mo.skin = player.mo.skin
		mo.spritexscale = player.mo.spritexscale
		mo.spriteyscale = player.mo.spriteyscale
		if not condition(player) then
			if time[#player] == nil then
				time[#player] = 0
			end
			time[#player] = time[#player] - 1
			if time[#player] <= 0 then
				player.mo[field] = nil -- we don't need to do this, but just in case
				P_RemoveMobj(mo)
				mo = nil
				time[#player] = nil
			end
		else
			if time[#player] == nil then
				time[#player] = 15
			end
			time[#player] = time[#player] + 1
			if time[#player] > 15 then
				time[#player] = 15
			end
		end
		if mo ~= nil then
			mo.frame = (player.mo.frame & FF_FRAMEMASK) | ((9 - time[#player] / 3) << FF_TRANSSHIFT)
		end
	elseif player.mo[field] == nil then
		if condition(player) then
			local mo = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_OVERLAY)
			P_SetMobjStateNF(mo, S_INVISIBLE) -- delay update with one tic to avoid triggering state bugs
			mo.destscale = mo.scale
			mo.angle = player.drawangle
			mo.pitch = player.mo.pitch
			mo.roll = player.mo.roll
			mo.spriteroll = player.mo.spriteroll
			mo.skin = player.mo.skin
			mo.color = SKINCOLOR_SUPERSKY1
			mo.colorized = true
			mo.target = player.mo
			mo.frame = (player.mo.frame & FF_FRAMEMASK) | FF_TRANS90
			player.mo[field] = mo
			S_StartSound(player.mo, sfx_monton)
			time[#player] = 0
		end
	end
end

addHook("PostThinkFrame", function ()
	for player in players.iterate do
		if not (player and player.mo) then
			return
		end
		local dest = player.mo.ht_trans
		if not dest then
			dest = 0
		end
		if ht_playertrans[#player] < dest then
			ht_playertrans[#player] = ht_playertrans[#player] + 1
		elseif ht_playertrans[#player] > dest then
			ht_playertrans[#player] = ht_playertrans[#player] - 1
		end
		local trans = ht_playertrans[#player] * 6 / MAXPLAYERTRANS
		player.mo.frame = (player.mo.frame & ~FF_TRANSMASK) | (trans << FF_TRANSSHIFT)
		if player.mo.scale ~= player.mo.destscale then
			playerscaling[#player] = true
			if not S_SoundPlaying(player.mo, sfx_s3kc6l) then
				S_StartSound(player.mo, sfx_s3kc6l)
			end
			player.mo.renderflags = player.mo.renderflags | RF_FULLBRIGHT
			player.mo.color = SKINCOLOR_SUPERSILVER1
			player.mo.colorized = true
			if leveltime % 5 == 0 then
				local mo = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SCALEGHOST)
				if trans then
					mo.state = S_SCALEGHOST1 + trans - 1
				end
				mo.destscale = mo.scale
				mo.angle = player.drawangle
				mo.pitch = player.mo.pitch
				mo.roll = player.mo.roll
				mo.spriteroll = player.mo.spriteroll
				mo.cusval = player.mo.sprite2
				mo.cvmem = (player.mo.frame & FF_FRAMEMASK)
				mo.skin = player.mo.skin
				mo.color = SKINCOLOR_SUPERSILVER1
				mo.colorized = true
			end
		elseif playerscaling[#player] then
			playerscaling[#player] = false
			S_StopSoundByID(player.mo, sfx_s3kc6l)
			player.mo.renderflags = player.mo.renderflags & ~RF_FULLBRIGHT
			player.mo.color = player.skincolor
			player.mo.colorized = false
		end

		HandlePlayerOverlay(player, "paceoverlay", paceframe, pacetime, SKINCOLOR_SUPERSKY1, function (player)
			return player.normalspeed > skins[player.mo.skin].normalspeed or player.jumpfactor > skins[player.mo.skin].jumpfactor
		end)

		if tpstate[#player] == 1 then
			player.mo.spritexscale = player.mo.spritexscale - FRACUNIT / 10
			player.mo.spriteyscale = (FRACUNIT - player.mo.spritexscale) + FRACUNIT
			if player.mo.spritexscale < 0 then
				player.mo.spritexscale = 0
				if player.mo.target and player.mo.target.valid then
					-- only finalize the teleport if the target still exists
					if CV_FindVar("tailspickup").value then
						P_SetOrigin(player.mo, player.mo.target.x, player.mo.target.y, player.mo.target.z)
					else
						-- adjust for player collision
						P_SetOrigin(player.mo, player.mo.target.x, player.mo.target.y, player.mo.target.z + player.mo.target.height)
					end
					player.mo.target = nil
				end
				S_StartSound(player.mo, sfx_s3k75)
				tpstate[#player] = 2
			end
		elseif tpstate[#player] == 2 then
			player.mo.spritexscale = player.mo.spritexscale + FRACUNIT / 10
			player.mo.spriteyscale = (FRACUNIT - player.mo.spritexscale) + FRACUNIT
			if player.mo.spritexscale >= FRACUNIT then
				player.mo.spritexscale = FRACUNIT
				player.mo.spriteyscale = FRACUNIT
				tpstate[#player] = nil
			end
		end

		if flipping[#player] then
			player.viewrollangle = player.viewrollangle / 2
			player.mo.spriteroll = player.mo.spriteroll / 2
			if player.mo.spriteroll == 0 then
				flipping[#player] = nil
			end
		end
	end
end)

function A_CopyPlayerState(mo)
	mo.sprite2 = mo.cusval
	mo.frame = mo.cvmem | (mo.frame & ~FF_FRAMEMASK)
end
